/*
 * Copyright (c) 2003, 2004 Jussi Koskela
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
package fi.iki.hsivonen.webcache.front;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;
import java.util.Timer;

import org.mortbay.http.HttpException;
import org.mortbay.http.HttpFields;
import org.mortbay.http.HttpRequest;
import org.mortbay.http.HttpResponse;
import org.mortbay.http.handler.AbstractHttpHandler;

public class CacheHandler extends AbstractHttpHandler{

    /** invalidation key, which invalidates the whole cache */
    private static final String INVALIDATE_ALL = "*";

    /** the url -> file map with LRU invalidation support */
    private LRUFileMap fileMap;

    /** maximum disk usage in bytes */
    private long maxDiskUsage;
    /** current disk usage in bytes */ 
    private long diskUsage;

    /** lock for disk usage */
    private Object diskUsageLock;

    /** cache directory */
    private File directory;

    /** invalidator timer */
    private Timer timer;
    /** map of expire timer tasks */
    private Map timerTaskMap;

    /** map of dependencies using key to file mapping*/
    private LinkedHashMap dependencyMap;
    /** map of dependencies using file to key mapping*/
    private LinkedHashSet recordInvSet;

    /** port number for client request */
    private int clientPort;
    /** port number for invalidation requests */
    private int invalidatorPort;

    /**
     * Class constructor.
     * @param directory the cache directory
     * @param maxDiskUsage the maximum disk usage in bytes
     * @param maxFiles the maximum number of cached files
     * @param clientPort the port number for client requests
     * @param invalidatorPort the port number for invalidation requests
     */
    public CacheHandler(File directory, long maxDiskUsage, int maxFiles, 
                        int clientPort, int invalidatorPort) {
        super();
        this.directory = directory;
        this.maxDiskUsage = maxDiskUsage;
        this.clientPort = clientPort;
        this.invalidatorPort = invalidatorPort;

        fileMap = new LRUFileMap(maxFiles);

        diskUsageLock = new Object();

        timer = new Timer();
        timerTaskMap = new LinkedHashMap();

        dependencyMap = new LinkedHashMap();
        recordInvSet = new LinkedHashSet();
    }

    /**
     * Returns the cache directory.
     * @return the cache directory
     */
    public File getDirectory() {
        return directory;
    }

    /**
     * Acquires the instance of the cached file with the given url.
     * @param url the url of the file
     * @return the instance of the cached file
     */
    public CachedFile acquireCachedFile(String url) throws IOException {
        FrontServer.debug("Acquire cached file " + url);
        CachedFile cf = null;

        synchronized (fileMap) {
            cf = (CachedFile)fileMap.get(url);
            if (cf == null) {
                cf = new CachedFile(this, url);
                fileMap.put(url, cf);
            }
            cf.acquire();
        }
        cf.delayedInit();
        return cf;
    }

    /**
     * Start recording invalidations for the given file.
     * @param cf the cached file
     */
    public void recordInvalidations(CachedFile cf) {
        FrontServer.debug("Record invalidations " + cf.getURL());
        synchronized (recordInvSet) {
            recordInvSet.add(cf);
        }
    }

    /**
     * Registers dependencies for the given file and stops recording
     * the dependencies as a side effect. 
     * @param cf the cached file
     * @param ds the dependency set
     */
    public void registerDependencies(CachedFile cf, Set ds) {
        FrontServer.debug("Register dependencies " + ds + " " + cf.getURL());
        synchronized (dependencyMap) {
            Iterator iter = ds.iterator();
            while (iter.hasNext()) {
                String key = (String)iter.next();
                
                Set dependencies = (Set)dependencyMap.get(key);
                if (dependencies == null) {
                    dependencies = new HashSet();
                    dependencyMap.put(key, dependencies);
                }
                dependencies.add(cf);
            }

            FrontServer.debug("Stop recording invalidations " + cf.getURL());
            synchronized (recordInvSet) {
                recordInvSet.remove(cf);
            }
        }
    }

    /**
     * Removes the dependencies for the given file.
     * @param cf the cached file
     * @param ds the dependency set
     */
    public void removeDependencies(CachedFile cf, Set ds) {
        FrontServer.debug("Remove dependencies " + ds + " " + cf.getURL());
        synchronized (dependencyMap) {
            Iterator iter = ds.iterator();
            while (iter.hasNext()) {
                String key = (String)iter.next();
                Set depFiles = (Set)dependencyMap.get(key);
		if (depFiles != null) {
		    depFiles.remove(cf);
		}
            }
        }

        synchronized (recordInvSet) {
            recordInvSet.remove(cf);
        }
    }


    /**
     * Invalidates the cached file with the given url.
     * @param url the url of the file
     */
    public void invalidateURL(String url) {
        FrontServer.debug("Invalidate URL " + url);
        CachedFile cf = null;
        synchronized (fileMap) {
            cf = (CachedFile)fileMap.remove(url);
        }

        if (cf != null) {
            cf.invalidate();
        }
    }

    /**
     * Invalidates all the cached files.
     */
    public void invalidateAll() {
        FrontServer.debug("Invalidate all");
        LinkedList workList = null;
        synchronized (fileMap) {
            workList = new LinkedList(fileMap.values());
            fileMap.clear();
        }
        
        Iterator iter = workList.iterator();
        while (iter.hasNext()) {
            CachedFile cf = (CachedFile)iter.next();
            cf.invalidate();
        }
    }

    /**
     * Registers an expiration for the given file.
     * @param cf the cached file
     * @param expires the date of expiration
     */
    public void registerExpiration(CachedFile cf, Date expires) {
        FrontServer.debug("Register expiration " + expires + " " + cf.getURL());
        ExpireTimerTask oldTask = null;
        synchronized (timerTaskMap) {
            oldTask = (ExpireTimerTask)timerTaskMap.get(cf);
        }
        if (oldTask != null) {
            if (expires.before(oldTask.getDate())) {
                oldTask.cancel();
            } else {
                return;
            }
        }
        ExpireTimerTask task = new ExpireTimerTask(expires, cf, this);
        
        synchronized (timerTaskMap) {
            timerTaskMap.put(cf, task);
        }
        timer.schedule(task, expires);
    }

    /**
     * Cancels the expiration of the given file.
     * @param cf the cached file
     */
    public void cancelExpiration(CachedFile cf) {
        ExpireTimerTask oldTask = null;
        synchronized (timerTaskMap) {
            oldTask = (ExpireTimerTask)timerTaskMap.remove(cf);
        }
        if (oldTask != null) {
            oldTask.cancel();
        }
    }

    /**
     * Handles the expiration of the given file.
     * @param cf the cached file
     */
    public void onExpire(CachedFile cf) {
        FrontServer.debug("On expire " + cf.getURL());
        synchronized (timerTaskMap) {
            timerTaskMap.remove(cf);
        }
        invalidateURL(cf.getURL());
    }


    /**
     * Requests disk space for the given number of bytes.
     * @param bytes the required disk space in bytes
     */
    public boolean requestDiskSpace(long bytes) {
        FrontServer.debug("Request disk space " + bytes + "(" + diskUsage + ")");
        LinkedList invalidateList = new LinkedList();
        long requiredSpace = 0;

        synchronized (diskUsageLock) {
            requiredSpace = diskUsage + bytes - maxDiskUsage;

            if (requiredSpace > 0) {
                synchronized (fileMap) {
                    Iterator iter = fileMap.values().iterator();
                    while (iter.hasNext() && (requiredSpace > 0)) {
                        CachedFile cf =(CachedFile)iter.next();
                        iter.remove();
                        
                        requiredSpace -= cf.getDiskUsage();
                        invalidateList.addLast(cf);
                    }
                }
            }
            diskUsage += bytes;
        }
        
        Iterator iter = invalidateList.iterator();
        while (iter.hasNext()) {
            CachedFile cf = (CachedFile)iter.next();
            cf.invalidate();
        }
                
        return true;
    }
    
    /**
     * Releases the given number of bytes of disk space.
     * @param bytes the released disk space in bytes
     */
    public void releaseDiskSpace(long bytes) {
        FrontServer.debug("Release disk space " + bytes);
        synchronized (diskUsageLock) {
            diskUsage -= bytes;
        }
    }

    /**
     * Handle the HTTP requests.
     */
    public void handle(String pathInContext, String pathParams,
                       HttpRequest request, HttpResponse response)
        throws HttpException, IOException {

        int port = request.getHttpConnection().getServerPort();

        if (port == clientPort) {
            handleClientRequest(request, response);
        } else if (port == invalidatorPort) {
            handleInvalidationRequest(request, response);
        }        
    }

    /**
     * Handle the invalidation requests.
     */
    public void handleInvalidationRequest(HttpRequest request, HttpResponse response) 
        throws HttpException, IOException {
        FrontServer.debug("Handle invalidation request");

        request.setHandled(true);
        
        InputStream in = request.getInputStream();
        InputStreamReader reader = new InputStreamReader(in);
        BufferedReader bufferedReader = new BufferedReader(reader);
        
        while (true) {
            String key = null;
            try {
                key = bufferedReader.readLine();
            } catch (IOException ioe) {
                FrontServer.debug("Failed to read line from invalidate post");
            }            
            if (key != null) {
                if (key.equals(INVALIDATE_ALL)) {
                    synchronized (dependencyMap) {
                        dependencyMap.clear();
                    }
                    invalidateAll();
                } else {
                    Set dependencies = null;
                    synchronized (dependencyMap) {
                        dependencies = (Set)dependencyMap.remove(key);
                    }
                    
                    if (dependencies != null) {
                        Iterator iter = dependencies.iterator();
                        while (iter.hasNext()) {
                            CachedFile cf = (CachedFile)iter.next();
                            invalidateURL(cf.getURL());
                        }
                    }
                    
                    synchronized (recordInvSet) {
                        Iterator iter = recordInvSet.iterator();
                        while (iter.hasNext()) {
                            CachedFile cf = (CachedFile)iter.next();
                            cf.addInvalidation(key);
                        }
                    }
                }
            } else {
                break;
            }
        }

        response.setStatus(HttpResponse.__204_No_Content);
        in.close();

    }

    /**
     * Handle the client requests.
     */
    public void handleClientRequest(HttpRequest request, HttpResponse response)
        throws HttpException, IOException {
        String url = request.getURI().toString();
        FrontServer.debug("Handle client request " + url);
        request.setHandled(true);
        
	CachedFile cf = null;
        try {
            if (!request.getMethod().equals(HttpRequest.__GET) &&
                !request.getMethod().equals(HttpRequest.__POST) &&
                !request.getMethod().equals(HttpRequest.__HEAD)) {
                
                response.setField(HttpFields.__Allow, "GET,POST,HEAD");
                response.setStatus(HttpResponse.__405_Method_Not_Allowed);
                System.err.println("Method not allowed " + url);
                return;
            }

            try {
                cf = acquireCachedFile(url);
            } catch (IOException ioe) {
                System.err.println(ioe);
                invalidateURL(url);
                response.setStatus(HttpResponse.__500_Internal_Server_Error);
                return;
            }

            FrontServer.debug("Got cached file " + url);

            
            int responseCode = cf.getResponseCode();
            String contentType = cf.getContentType();
            int contentLength = cf.getContentLength();
            long lifeTime = cf.getLifeTime();
            long lastModified = cf.getLastModified();


            if ((responseCode == HttpResponse.__302_Moved_Temporarily) ||
                (responseCode == HttpResponse.__301_Moved_Permanently)) {
                response.sendRedirect(cf.getRedirectLocation());
                FrontServer.debug("Redirect to " + cf.getRedirectLocation() + 
                                  " from " + url);
                return;
            }
            
            long ifUnmodifiedSince = 
                request.getDateField(HttpFields.__IfUnmodifiedSince);        
            if ((ifUnmodifiedSince > 0) && (lastModified > ifUnmodifiedSince)){
                response.setStatus(HttpResponse.__412_Precondition_Failed);
                FrontServer.debug("Precondition failed for " + url);
                return;
            }
        
            long ifModifiedSince = 
                request.getDateField(HttpFields.__IfModifiedSince);

            if ((ifModifiedSince > 0) && (lastModified <= ifModifiedSince)) {
                response.setStatus(HttpResponse.__304_Not_Modified);
                FrontServer.debug("Not Modified " + url);
                return;
            }

            response.setStatus(responseCode);

            // initialize the header
            if (contentType != null) {
                response.setContentType(contentType);
            }
        
            if (contentLength != -1) {
                response.setContentLength(contentLength);
            }

            if (lifeTime != -1) {
		response.setDateField(HttpFields.__Expires, 
                                      System.currentTimeMillis() + 
                                      1000L * lifeTime);
            } else {
                response.setField(HttpFields.__Expires, 
                                  HttpFields.__01Jan1970);
            }
        
            response.setDateField(HttpFields.__LastModified,
                                  new Date(lastModified));

            // should we send only the head?
            if (request.getMethod().equals(HttpRequest.__HEAD)) {
                FrontServer.debug("Header sent for " + url);
                return;
            }
            
            FrontServer.debug("Begin sending " + url);

            InputStream in = null;
            OutputStream out = null;
            try {
                int bufferSize = 8192;
                byte buffer[] = new byte[bufferSize];

                in = cf.getInputStream();
                out = response.getOutputStream();
                
                int pos = 0;
                
                while (cf.waitInput(pos)) {
                    int read = in.read(buffer, 0, bufferSize);
                    if (read != -1) {
                        out.write(buffer, 0, read);
                        pos += read;
                    }
                    FrontServer.debug("Sending data " + pos + " " + url);
                }
            } catch (IOException ioe) {
                System.err.println(ioe);
                // Jetty has terminated the socket
            } finally {
                try {
                    if (in != null) {
                        in.close();
                    }
                } catch (IOException ioe) {}
                try {
                    if (out != null) {
                        out.close();
                    }
                } catch (IOException ioe) {}
                response.commit();
                FrontServer.debug("Response sent for " + url);
            }
        } finally {
            if (cf != null) {
                cf.release();
            }
        }
    }
}
